""" ddeint/__init__.py """

# Credits to Zulko for this module
# https://github.com/Zulko/ddeint
# Some tweaks habe been made for compatibility reasons
__all__ = ["ddeint"]

from .ddeint import ddeint

from .version import __version__



import numpy as np
from matplotlib import pyplot as plt
from matplotlib import animation, rc

rc("text", usetex=True)


def animate(sol, time=None,
            fig=None, ax=None, title="", savepath=None, fig_size=(6, 6), variable=0, go_by=1, **kwargs):
    """
    Method that animates the time evolution of the defined system. Optionally saves an mp4 video file to
    'savepath'.

    Parameters
    ----------
    sol : numpy.array
        Array of shape either (#Variables, #Timesteps, #Dim1) or (#Variables, #Timesteps, #Dim1, #Dim2)
    time : numpy.array
        (Optional, default=None) Array of time stamps, if None defaults to 0 + 1 per time step.
    title : string
        (Optional, default="") Give a title for the resulting animation
    savepath : string
        (Optional, default=None) Give a savepath for the animation
    fig_size : tuple
        (Optional, default=(6, 6)) Give a figure size for matplotlib
    variable : int
        (Optional, default=0) Gives the index of the variable to show (if there are more than one)
    go_by : int
        (Optional, default=1) Steps through the solution by `go_by` number of steps
    kwargs
        Additional key word arguments are passed to matplotlib.animation.FuncAnimation

    Returns
    -------
    matplotlib.animation.FuncAnimation

    """
    if fig is None or ax is None:
        fig, ax = plt.subplots(figsize=fig_size)

    if time is None:
        time = range(sol.shape[1])

    dimension = len(sol.shape) - 2

    if dimension == 2:
        sol_img = ax.imshow(sol[variable, 0])
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        colorbar = plt.colorbar(sol_img)
    elif dimension == 1:
        sol_img, = ax.plot(sol[variable, 0])
        ax.set_ylim(np.min(sol[variable, :]), np.max(sol[variable, :]))
        ax.set_xlabel("x")
        ax.set_ylabel("u")
    else:
        raise NotImplementedError("Only 1D and 2D solutions can be animated.")

    # This feels stupid (trying to avoid "if" in animation call)+
    def anime2d(i):
        ax.set_title(title + f" Time: {time[i]:.3f}")
        data_max = np.max(sol[variable, i])
        data_min = np.min(sol[variable, i])

        sol_img.set_data(sol[variable, i])
        sol_img.set_clim(data_min, data_max)


        return sol_img

    def anime1d(i):
        ax.set_title(title + f" Time: {time[i]:.3f}")

        sol_img.set_ydata(sol[variable, i])
        return sol_img

    # This feels even more stupid
    if dimension == 2:
        anime = anime2d
    elif dimension == 1:
        anime = anime1d

    frames = np.arange(0, len(time), go_by)
    anima = animation.FuncAnimation(fig, anime, frames=frames, repeat=False, **kwargs)
    if savepath is not None:
        Writer = animation.writers["ffmpeg"]
        writer = Writer(fps=15, bitrate=1800)
        anima.save(savepath, writer=writer)
    # else:
    # plt.show()

    return anima


def barplot_parameters(sigma, sigma_ref=None, labels=None, width=0.35, table_fontsize=None, **kwargs):
    """
    Method that visualizes a set of parameters as a bar plot and may compare it to the true set of parameters

    Parameters
    ----------
    sigma : numpy.array
        Array of shape (# Varibales, # library terms), essentially `model_estimation.model.sigma`
    sigma_ref : numpy.array, default=None
        Array of same shape as sigma with reference values.
    labels : list of strings, default=None
        List of strings to label the x-axis with. If none indices are used
    width : float, default=0.35
        Width of bars in barplot
    table_fontsize : float, default=None
        Overwrites the automatic fontsize of the table which shows library labels
    kwargs
        Additional Keyword arguments are passed on to matplotlib.pyplot.subplot
    """
    plt.rcParams.update({"text.usetex": False})

    ind = np.arange(sigma.shape[1])
    labels_loc = range(sigma.shape[1])
    if labels is None:
        labels = labels_loc

    ratios = [2.5] + [1]
    # fig, ax = plt.subplots(nrows=sigma.shape[0], ncols=2, sharex=True, gridspec_kw={'width_ratios': ratios}, **kwargs)

    # only works for matplotlib >= 3.6.0:
    # fig, ax = plt.subplots(nrows=sigma.shape[0], ncols=2, sharex=True, width_ratios=ratios, **kwargs)

    fig = plt.figure(**kwargs)  # layout="constrained"
    spec = fig.add_gridspec(sigma.shape[0], 2, width_ratios=ratios)
    ax = []

    for i in range(sigma.shape[0]):
        axis = fig.add_subplot(spec[i, 0])
        rect = axis.bar(ind, sigma[i], width, color="tab:blue")
        leg = [(rect[0],), ("Estimated",)]
        axis.set_axisbelow(True)

        if sigma_ref is not None:
            rect_true = axis.bar(ind + width, sigma_ref[i], width, color="tab:purple")
            leg = [(rect[0], rect_true[0]), ("Estimated", "Reference")]

        axis.set_ylabel(f"Value for Variable {i}")
        axis.set_xticks(ind + (sigma_ref is not None) * (width / 2))
        axis.set_xticklabels(labels_loc)
        axis.legend(leg[0], leg[1])
        axis.grid()
        ax.append(axis)

    ax_table = fig.add_subplot(spec[:, 1])
    ax_table.axis("off")
    text = np.array([labels_loc, labels])
    table = ax_table.table(cellText=text.T, colLabels=("Index", "Label"), loc="center", colWidths=[0.25, 1])
    if table_fontsize is not None:
        table.auto_set_font_size(False)
        table.set_fontsize(table_fontsize)
    plt.tight_layout()

    plt.show()
    plt.rcParams.update({"text.usetex": True})



from tsme.time_simulation.simulator import AbstractTimeSimulator, AbstractDiscreteTimeSimulator
import numpy as np


class Lorenz(AbstractTimeSimulator):
    def __init__(self, ic=None, params=None):
        if params is None:
            params = [1.0, 1.0, 1.0]
        super().__init__(ic)
        self.sigma = params[0]
        self.rho = params[1]
        self.beta = params[2]

    def rhs(self, t, u_in):
        """
        Right-hand side function of the Lorenz system.

        Parameters
        ----------
        t : float
            current time value during simulation (here only included to interface with scipy's IVP solver
        u_in : numpy.array
            current state as array of shape (3,) for variables :math:`x`, :math:`y`, :math:`z`

        Notes
        -----
        The lorenz system consists of the following coupled ordinary differential equations:

        .. math::

            \\frac{\\text{d}x}{\\text{d}t} &= \\sigma \\left(y - x \\right) \\\\
            \\frac{\\text{d}y}{\\text{d}t} &= x \\left(\\rho - z \\right) - y \\\\
            \\frac{\\text{d}z}{\\text{d}t} &= x\\, y - \\beta \\, z

        The corresponding parameters are `self.sigma` (:math:`\\sigma`, default: 1.0), `self.rho`
        (:math:`\\rho`, default:1.0) and `self.beta` (:math:`\\beta`, default: 1.0). These are set during initialisation
         using a list in the keyword argument `params`.

        """
        x, y, z = u_in

        x_next = self.sigma * (y - x)
        y_next = x * (self.rho - z) - y
        z_next = x * y - self.beta * z

        return np.array([x_next, y_next, z_next])


class Burgers(AbstractTimeSimulator):
    def __init__(self, ic, dom, bc="periodic", diff="finite_difference"):
        super().__init__(ic, domain=dom, bc=bc, diff=diff)

    def rhs(self, t, u_in):
        """
        Burgers Equation in 1D.

        Parameters
        ----------
        t : float
            current time value during simulation (here only included to interface with scipy's IVP solver
        u_in : numpy.array
            current state as array of shape (1, N) for variable :math:`u(x)`, where N is the number of spatial
            discretization steps

        Notes
        -----
        The burgers equation in 1D reads here:

        .. math::

            \\frac{\\text{d}u}{\\text{d}t} = \\frac{\\text{d}^2u}{\\text{d}x} - u\\, \\frac{\\text{d}u}{\\text{d}x}

        """
        u = u_in[0]
        u_next = self.diff.d_dx(u, 2) - u * self.diff.d_dx(u, 1)

        return np.array([u_next])


class CahnHilliard(AbstractTimeSimulator):
    def __init__(self, ic, dom, params=None, bc="periodic", diff="finite_difference"):
        if params is None:
            params = [1.0, 1.0]
        super().__init__(ic, domain=dom, bc=bc, diff=diff)
        self.D = params[0]
        self.alpha = params[1]

    def rhs(self, t, u_in):
        """
        Cahn-Hilliard Equation

        Parameters
        ----------
        t : float
            current time value during simulation (here only included to interface with scipy's IVP solver
        u_in : numpy.array
            current state as array of shape (1, N, N) for variable :math:`u(x, y)`, where N is the number of spatial
            discretization steps

        Notes
        -----
        The Cahn-Hilliard equation reads here:

        .. math::

            \\frac{\\text{d}u}{\\text{d}t} = \\nabla^2 \\left(u^3 - \\alpha \\, u - D \\, \\nabla^2 u \\right)

        The corresponding parameters are `self.D` (:math:`D`, default: 1.0) and
        `self.alpha` (:math:`\\alpha`, default: 1.0). These are set during initialisation using a list in the keyword
        argument `params`.

        """
        u = u_in[0]
        f = u ** 3 - self.alpha * u - self.D * (self.diff.d_dx(u, 2) + self.diff.d_dy(u, 2))
        u_next = self.diff.d_dx(f, 2) + self.diff.d_dy(f, 2)

        return np.array([u_next])


class FitzHughNagumo(AbstractTimeSimulator):
    def __init__(self, ic, dom, params=None, bc="periodic", diff="finite_difference"):
        if params is None:
            params = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        super().__init__(ic, domain=dom, bc=bc, diff=diff)
        self.D_u = params[0]
        self.D_v = params[1]
        self.lamb = params[2]
        self.omega = params[3]
        self.kappa = params[4]
        self.tau = params[5]

    def rhs(self, t, u_in):
        """
        FitzHugh-Nagumo equation

        Parameters
        ----------
        t : float
            current time value during simulation (here only included to interface with scipy's IVP solver
        u_in : numpy.array
            current state as array of shape (2, N, N) for variables :math:`u(x, y)` and :math:`v(x, y)` where N is the
            number of spatial discretization steps

        Notes
        -----
        The FitzHugh-Nagumo equation reads here:

        .. math::

            f &= \\lambda \\, u - u^3 - \\kappa \\\\
            \\frac{\\text{d}u}{\\text{d}t} &= D_u^2 \\, \\nabla^2 u + f - \\omega \\, v \\\\
            \\tau \\, \\frac{\\text{d}v}{\\text{d}t} &= D_v^2 \\, \\nabla^2 v + u - v

        The corresponding parameters are: `self.D_u` (:math:`D_u`, default: 1.0), `self.D_v` (:math:`D_v`, default: 1.0)
        , `self.lamb` (:math:`\\lambda`, default: 1.0), `self.omega` (:math:`\\omega`, default: 1.0),
        `self.kappa` (:math:`\\kappa`, default: 1.0), `self.tau` (:math:`\\tau`, default: 1.0)

        """
        u = u_in[0]
        v = u_in[1]
        f = self.lamb * u - u ** 3 - self.kappa
        u_next = self.D_u ** 2 * (self.diff.d_dx(u, 2) + self.diff.d_dy(u, 2)) + f - self.omega * v
        v_next = self.D_v ** 2 * (self.diff.d_dx(v, 2) + self.diff.d_dy(v, 2)) + u - v

        return np.array([u_next, self.tau ** (-1) * v_next])


class KortewegDeVries(AbstractTimeSimulator):
    def __init__(self, ic, dom, bc="periodic", diff="finite_difference"):
        super().__init__(ic, domain=dom, bc=bc, diff=diff)

    def rhs(self, t, u_in):
        """
        Korteweg de Vries equation in 1D

        Parameters
        ----------
        t : float
            current time value during simulation (here only included to interface with scipy's IVP solver
        u_in : numpy.array
            current state as array of shape (1, N) for variable :math:`u(x)`, where N is the number of spatial
            discretization steps

        Notes
        -----
        The Kortweg de Vries equation reads here:

        .. math::

            \\frac{\\text{d}u}{\\text{d}t} = \\frac{\\text{d}^3u}{\\text{d}x^3} - 6 u\\, \\frac{\\text{d}u}{\\text{d}x}


        """
        u = u_in[0]
        u_next = - self.diff.d_dx(u, 3) - 6 * u * self.diff.d_dx(u, 1)

        return np.array([u_next])


class HenonMap(AbstractDiscreteTimeSimulator):
    def __init__(self, ic, params=None):
        if params is None:
            params = [0.1, 0.1, 0.01]
        super().__init__(ic)
        self.a = params[0]
        self.b = params[1]
        self.gamma = params[2]

    def map(self, u, i):
        """
        Henon map

        Parameters
        ----------
        u : numpy.array
            Array containing system states over time, i.e. of shape (# total time steps, # variables)
        i : integer
            current time step

        Returns
        -------
        numpy.array
            u with a set entry for i+1

        Notes
        -----
        Here the Henon map reads:

        .. math::

            x_{i+1} &= 1 - a\\, x_i^2 + y_i + \\gamma \\cdot \\text{noise} \\cdot x_i \\\\
            y_{i+1} &= b\\, x_i

        The parameters are given by the respective attributes `self.a` (:math:`a`, default: 0.1),
        `self.b` (:math:`b`, default: 0.1) and `self.gamma` (:math:`\\gamma`, default: 0.01)

        """
        x_i = u[i][0]
        y_i = u[i][1]

        x_i_1 = 1 - self.a * x_i ** 2 + y_i + self.gamma * np.random.normal() * x_i
        y_i_1 = self.b * x_i

        u[i + 1] = [x_i_1, y_i_1]

        return u


import numpy as np
from scipy.integrate import solve_ivp
from .differentiation import DifferentialOperator, get_differential_operator
from .ddeint import ddeint


class AbstractTimeSimulator:
    """
    This abstract class provides basic time simulation functionality for ODEs and PDEs (1D or 2D) using Scipys initial
    value solver. It is intended to be extended with needed attributes and a right-hand-side function (rhs), It also
    provides methods for handling spatial differentiation and boundary conditions.
    """

    def __init__(self, ic, domain=None, bc=None, diff=None, **kwargs):
        """
        Basic constructor method. Provides spatial coordinates (if applicable) and records basic information of the
        problem. Also creates some protected attributes for differentiation methods.

        Parameters
        ----------
        ic : array-like
            array of initial conditions, should be in the shape (# Variables, ) for ODEs, (# Variables, # First
            spatial dimension) for 1D PDEs and (# Variables, # First spatial dimension, # Second spatial dimension) for
            2D PDEs
        domain : tuple
            physical domain of the problem, for ODE value is None (default), for PDE shape should be ((x_min, x_max), )
            in the 1D case and ((x_min, x_max), (y_min, y_max)) for 2D (3D case analogous)
        bc : string
            Set the boundary conditions in case of spatially extended system, otherwise 'None' (default). Options
            include "periodic" or "neumann"
        diff : string
            Set the differentiation method in case of spatially extended system, otherwise 'None' (default). Options
            include "finit_difference" or "pseudo_spectral"

        Attributes
        ----------
        To be extended
        """
        if diff == "pseudo_spectral" and bc != "periodic":
            bc = "periodic"
            print("Warning: Boundary condition forced to 'periodic' due to pseudo-spectral method!")

        assert diff is None or diff == "pseudo_spectral" or diff == "finite_difference", \
            "Unsupported differentiation method."

        self.n_variables = ic.shape[0]  # Number of variables of the problem
        self.domain = domain  # Passed domain as specified above
        self.initial_state = ic  # Passed initial conditions as stated above
        self._dimensions = None  # number of spatial dimension of the problem
        self._discret_x = None  # spatial discretization of the problem in first dimension (if applicable)
        self._discret_y = None  # spatial discretization of the problem in second dimension (if applicable)
        self.diff = None  # instance of DiffirentialOperator handling spatial differentiation
        self.solver = None  # to save the scipy solver object (maybe remove)
        self.sol = None  # to save time simulation
        self.sol_dot = None  # to save derivative of time simulation
        self.time = None  # time stamps of the simulation
        self.d_dt = None  # to save (finite difference) differential operator w.r.t. time

        # Check provided domain to see for different use-cases: ODE, 1D PDE, 2D PDE
        if self.domain is None:
            self._mode = "ODE"
        else:
            self._mode = "PDE"
            self._dimensions = len(domain)
            self._discret_x = self.initial_state.shape[1]

            # Check to see if discretization along the dimensions match and also if domain matches initial conditions
            if self._dimensions >= 2:
                try:
                    self.initial_state.shape[2]
                except IndexError:
                    raise ValueError("Dimension of domain and initial conditions do not match.")
                # assert self._discret == self.initial_state.shape[2], "Discretization needs to be the same in either " \
                #                                               "dimension."
                self._discret_y = self.initial_state.shape[2]

            if self._dimensions == 3:
                try:
                    self.initial_state.shape[3]
                except IndexError:
                    raise ValueError("Dimension of domain and initial conditions do not match.")
                # assert self._discret == self.initial_state.shape[2], "Discretization needs to be the same in either " \
                #                                               "dimension."
                self._discret_z = self.initial_state.shape[3]

        # All of the following enables use of spatial coordinates and derivatives in rhs function call
        if self._dimensions is None:
            pass
        elif self._dimensions == 1:
            self.Lx = self.domain[0][1] - self.domain[0][0]  # Absolute length of the spatial interval
            self.x = np.linspace(self.domain[0][0], self.domain[0][1], num=self._discret_x)  # spatial coordinates
            # create object handling spatial derivatives
            if "x" in kwargs:
                self.x = kwargs.get("x")
            self.diff = DifferentialOperator({"Lx": self.Lx, "x": self.x, "nx": self._discret_x},
                                             boundary=bc, method=diff)
        elif self._dimensions == 2:
            self.Lx = self.domain[0][1] - self.domain[0][0]  # Absolute length of the first spatial interval
            self.Ly = self.domain[1][1] - self.domain[1][0]  # Absolute length of the second spatial interval
            self.x = np.linspace(self.domain[0][0], self.domain[0][1], self._discret_x)  # spatial coordinates
            self.y = np.linspace(self.domain[1][0], self.domain[1][1], self._discret_y)
            if "x" in kwargs:
                self.x = kwargs.get("x")
            if "y" in kwargs:
                self.y = kwargs.get("y")
            # self.x_mesh, self.y_mesh = np.meshgrid(self.x, self.y)
            # create object handling spatial derivatives
            # TODO: Remove discretization from function call
            self.diff = DifferentialOperator({"Lx": self.Lx, "x": self.x,
                                              "Ly": self.Ly, "y": self.y, "nx": self._discret_x, "ny": self._discret_y},
                                             boundary=bc, method=diff)
        elif self._dimensions == 3:
            self.Lx = self.domain[0][1] - self.domain[0][0]  # Absolute length of the first spatial interval
            self.Ly = self.domain[1][1] - self.domain[1][0]  # Absolute length of the second spatial interval
            self.Lz = self.domain[2][1] - self.domain[2][0]  # Absolute length of the third spatial interval
            self.x = np.linspace(self.domain[0][0], self.domain[0][1], self._discret_x)  # spatial coordinates
            self.y = np.linspace(self.domain[1][0], self.domain[1][1], self._discret_y)
            self.z = np.linspace(self.domain[1][0], self.domain[1][1], self._discret_z)

            if "x" in kwargs:
                self.x = kwargs.get("x")
            if "y" in kwargs:
                self.y = kwargs.get("y")
            if "z" in kwargs:
                self.z = kwargs.get("z")
            # self.x_mesh, self.y_mesh = np.meshgrid(self.x, self.y)
            # create object handling spatial derivatives
            # TODO: Remove discretization from function call
            self.diff = DifferentialOperator({"Lx": self.Lx, "x": self.x,
                                              "Ly": self.Ly, "y": self.y,
                                              "Lz": self.Lz, "z": self.z,
                                              "nx": self._discret_x, "ny": self._discret_y, "nz": self._discret_z},
                                             boundary=bc, method=diff)
        else:
            raise NotImplementedError

    def rhs(self, t, u):
        """
        Abstract method for right-hand-side function of initial value problem. Per default available are:
        self.x (and self.y): spatial coordinates
        self.diff.d_dx(u, o): spatial derivative in first spatial component of order o
        self.diff.d_dy(u, o): spatial derivative in second spatial component of order o (if applicable)
        Other quantities should be computed in the constructor method when they are independent of u and t.

        Parameters
        ----------
        t : float
            current time during simulation
        u : array-like
            current (unflattened) variables during time simulation

        Returns
        -------
        u_next : array-like
            values of variables for the next time step
        """
        pass

    def _flat_rhs(self, t, uin):
        """
        Helper method that corrects input and output dimension between rhs function and scipys ivp solver
        Parameters
        ----------
        self
        t : float
            time argument to pass along to the function
        uin : array-like
            variable values to pass along to the function

        Returns
        -------
        flat_u_next : array-like
            a flattend version of the values for the variables in the next time step
            """
        if self._dimensions is None:
            u = uin
        elif self._dimensions == 1:
            u = np.reshape(uin, (self.n_variables, self._discret_x))
        elif self._dimensions == 2:
            u = np.reshape(uin, (self.n_variables, self._discret_x, self._discret_y))
        elif self._dimensions == 3:
            u = np.reshape(uin, (self.n_variables, self._discret_x, self._discret_y, self._discret_z))
        else:
            raise NotImplementedError

        result = self.rhs(t, u)

        return result.flatten()

    def simulate(self, time_interval, method="DOP853", t_eval=None):
        """
        Solve the initial value problem posed by the right-hand-side function and the initial condition (the latter
        being given during initialization).

        Parameters
        ----------
        time_interval : list
            Two-element list giving initial and final time
        method : string, default="DOP853"
            Passed to scipy's IVP solver to determine the integration method
        t_eval : array_like, default=None
            Also passed to scipy, samples the solution to give time stamps

        """
        # solve initial value problem
        self.solver = solve_ivp(self._flat_rhs, time_interval, self.initial_state.flatten(),
                                method=method, t_eval=t_eval)
        # save solution
        self.sol = [self.solver.y[:, i] for i in range(len(self.solver.y[0, :]))]
        # de-flatten output
        if self._dimensions is None:
            self.sol = np.array([np.reshape(item, (self.n_variables,)) for item in self.sol])
        elif self._dimensions == 1:
            self.sol = [np.reshape(item, (self.n_variables, self._discret_x)) for item in self.sol]
            self.sol = np.array(self.sol)
        elif self._dimensions == 2:
            self.sol = [np.reshape(item, (self.n_variables, self._discret_x, self._discret_y)) for item in self.sol]
            self.sol = np.array(self.sol)
        elif self._dimensions == 3:
            self.sol = [np.reshape(item, (self.n_variables, self._discret_x, self._discret_y, self._discret_z))
                        for item in self.sol]
            self.sol = np.array(self.sol)

        self.sol = np.array([self.sol[:, i] for i in range(self.sol.shape[1])])
        self.time = self.solver.t

        return self.sol

    def get_sol_dot(self, t=None, s=None):
        """
        Helper method that computes the time derivative for given time and solution. (This should approximate the value
        of the right-hand side numerically.)

        Parameters
        ----------
        t : array_like, default=None
            Array of time stamps (if None: use time stored in attributes)
        s : array_like, defautl=None
            Array of system states for every time step (if None: use solution from attributes)

        Returns
        -------
        numpy.array
            Numerical time derivative (finite-differences) of solution array

        """
        if t is None:
            time = self.time
        else:
            time = t
        if s is None:
            sol = self.sol
        else:
            sol = s

        assert time is not None, "Need to pass array of time stamps or perform a time simulation"
        assert sol is not None, "A solution array needs to be provided or a time simulation needs to be performed"

        self.d_dt = get_differential_operator(time, 1)
        self.sol_dot = self.d_dt(sol)[:, :-1]
        return self.sol_dot


class AbstractDiscreteTimeSimulator:
    """
    This abstract class provides basic time simulation functionality for discrete maps. It is intended to be extended
    with needed attributes and a map function.
    """
    def __init__(self, ic):
        """
        Basic constructor method

        Parameters
        ----------
        ic : array_like
            Initial condition of shape (# Variables, )
        """
        self.init = ic
        self.time = None
        self.sol = None

    def map(self, u, i):
        """
        Abstract mapping method, needs to be implemented.

        Parameters
        ----------
        u : numpy.array
            Array containing system states over time, i.e. of shape (# total time steps, # variables)
        i : integer
            current time step

        Returns
        -------
        numpy.array
            u with a set entry for i+1

        """
        pass

    def simulate(self, steps):
        """
        Evolve the map for `steps` number of time steps

        Parameters
        ----------
        steps : integer
            Number of time steps for which to apply the discrete map

        Returns
        -------
        numpy.array
            Array of shape (# total number of steps, # variables)

        """
        self.sol = np.zeros((steps+1, *self.init.shape))
        self.sol[0] = self.init

        for step in range(steps):
            self.sol = self.map(self.sol, step)

        self.sol = self.sol.T
        return self.sol


class AbstractTimeDelaySimulator(AbstractTimeSimulator):
    """
    This abstract class is based on the 'AbstractTimeSimulator' class and provides basic time simulation functionality
    for ODEs and PDEs with time delay using a slightly modified version of the Scipy-based DDE integration package
    'ddeint'. It is intended to be extended with needed attributes and a delay model (in self.model). As of now only
    one fixed delayed variable is provided to the model (may be changed in the future).
    """
    def __init__(self, delay, domain=None, bc=None, diff=None):
        """
        Basic constructor method. It should be noted that for construction a "values_before_zero" function needs to be
        implemented, which for t=0 should return the initial state with the correct dimension of the problem.

        Parameters
        ----------
        delay : float
            Time delay
        domain : array_like
            Physical domain of the problem, for ODE value is None (default), for PDE shape should be ((x_min, x_max), )
            in the 1D case and ((x_min, x_max), (y_min, y_max)) for 2D and analogous for the 3D case
        bc : string
            Set the boundary conditions in case of spatially extended system, otherwise 'None' (default). Options
            include "periodic" or "neumann"
        diff : string
            Set the differentiation method in case of spatially extended system, otherwise 'None' (default). Options
            include "finit_difference" or "pseudo_spectral"
        """
        ic = self.values_before_zero(0)
        super().__init__(ic, domain=domain, bc=bc, diff=diff)
        self.delay = delay

    def values_before_zero(self, t):
        """
        Function that needs to be implemented by the user. It should return values for the system state for t <= 0.
        Take special note that the shape of the output (#variable, # spatial_discretization) for PDEs or (#variables,)
        for ODEs should match the (user-implemented) model function.

        Parameters
        ----------
        t : float
            Time

        Returns
        -------
        numpy.ndarray
            System state at time t (<= 0)
        """
        pass

    def model(self, u, u_d, t):
        """
        Function that needs to be implemented by the user. Should return a new system sate with shape (#variable,
        #spatial_discretization) for PDEs or (#variables,) for ODEs

        Parameters
        ----------
        u : numpy.ndarray
            System state at current time t
        u_d : numpy.ndarray
            System state at time (t - self.delay)
        t : float
            Current time t

        Returns
        -------
        numpy.ndarray
            New system state
        """
        pass

    def _model_wrapper(self, Y, t, d):
        """
        Internal helper function that wraps the model function to pass it along to ddeint for simulation. Note that
        here we more or less define what is available to the model function, i.e. here we restrict the model to only one
        type of delay.

        Parameters
        ----------
        Y : ddeint.sate (not accurate)
            State as recorded by ddeint
        t : float
            Current time t
        d : float
            Delay for the model

        Returns
        -------
        numpy.ndarray
            New system state, but flattened
        """
        u = self.extract_Y(Y, t, 0)
        u_d = self.extract_Y(Y, t, d)

        du = self.model(u, u_d, t)

        return du.flatten()

    def _values_before_zero_wrapper(self, t):
        """
        Internal helper function that wraps the 'values_before_zero' function to accommodate different shapes, i.e.
        allow for PDEs

        Parameters
        ----------
        t : float
            Time t (<= 0)

        Returns
        -------
        numpy.ndarray
            System state at time t (<= 0), but flattened
        """
        return self.values_before_zero(t).flatten()

    def extract_Y(self, Y, t, d):
        """
        Helper function that takes a system state as it is recorded by ddeint, and returns the system state at time
        (t - d) in appropriate dimensions, i.e. unflattened.

        Parameters
        ----------
        Y : ddeint.sate (not accurate)
            System state as it is recorded by ddeint
        t : float
            Time
        d : float
            Delay

        Returns
        -------
        numpy.ndarray
            System state at time t (t - d) in appropriate dimensions, i.e. unflattened.
        """
        if self._dimensions is None:
            u = Y(t - d)
        elif self._dimensions == 1:
            u = np.reshape(Y(t - d), (self.n_variables, self._discret_x))
        elif self._dimensions == 2:
            u = np.reshape(Y(t - d), (self.n_variables, self._discret_x, self._discret_y))
        elif self._dimensions == 3:
            u = np.reshape(Y(t - d), (self.n_variables, self._discret_x, self._discret_y, self._discret_z))
        else:
            raise NotImplementedError
        return u

    def simulate_delay_model(self, t, d=None):
        """
        Use 'ddeint' to simulate the delay model in time.

        Parameters
        ----------
        t : array-like
            Time steps
        d : float, optional
            To overwrite the system delay (default: 'None').

        Returns
        -------
        numpy.ndarray
            Time series of the simulated system of shape (#variables, #time_steps, #spatial_discretization_1), where
            the number of discretization steps is either NA for ODEs or repeated per spatial dimension for PDEs.
        """
        if d is None:
            d = self.delay
        self.sol = ddeint(self._model_wrapper, self._values_before_zero_wrapper, t, fargs=(d,))
        # de-flatten output
        if self._dimensions is None:
            self.sol = np.array([np.reshape(item, (self.n_variables,)) for item in self.sol])
        elif self._dimensions == 1:
            self.sol = [np.reshape(item, (self.n_variables, self._discret_x)) for item in self.sol]
            self.sol = np.array(self.sol)
        elif self._dimensions == 2:
            self.sol = [np.reshape(item, (self.n_variables, self._discret_x, self._discret_y)) for item in self.sol]
            self.sol = np.array(self.sol)
        elif self._dimensions == 3:
            self.sol = [np.reshape(item, (self.n_variables, self._discret_x, self._discret_y, self._discret_z))
                        for item in self.sol]
            self.sol = np.array(self.sol)

        self.sol = np.array([self.sol[:, i] for i in range(self.sol.shape[1])])
        self.time = t

        return self.sol


import numpy as np
import scipy.integrate as integrate
from scipy.integrate import cumulative_trapezoid as cumtrapz
import optuna
from optuna.samplers import TPESampler

from gradient_free_optimizers import HillClimbingOptimizer
import nevergrad as ng
from hyperopt import fmin, hp, tpe
from tqdm import tqdm


def _assert_no_intersections(sets_list):
    """
    Assert that there are no intersections between all sets in the list.

    """
    # Flatten all the sets into one union to check if there's overlap
    all_elements = set()

    for s in sets_list:
        # Check if there's already any overlap with the current set
        assert all_elements.isdisjoint(s), f"Intersection found between {all_elements} and {s}"
        # Update all_elements with the current set
        all_elements.update(s)


# Maybe at some point add option for sklearns ridge regressor:
# from sklearn.linear_model import Ridge

# idea for alternative thresholding:
# do until included variables don't change or max_depth is reached:
#     Perform ordinary least square with all included variables
#     for all variables do:
#       remove variable and repeat the least squares
#       check BIC/integration error between the two least squares
#       if removing the variable made the error bigger than some threshold exclude the variable

def sequencing_threshold_ridge(lib, data, lamb, threshold=1e-4, max_it=10, norm=2, constraints=None,
                               scale_threshold=False, subset_threshold=None):
    """
    This reproduces the sequencing threshold ridge regression as it is implemented in PDE-FIND.

    Parameters
    ----------
    lib : np.array
        Time series of library terms as produced by a simulator instance via method create_library_time_series or of
        shape (# Variables, # Time Steps, # Library Terms, Dim1, Dim2(if applicable))
    data : np.array
        Time series to be fitted to (in shape (# Variable, # Time steps, Dim1, Dim2(if applicabe)))
    lamb : float
        Sparsity knob to promote sparsity in regularisation
    threshold : float or list-like
        (Optional, default: 1e-4) Threshold below which values are cut in sequence of regression
    max_it : int
        (Optional, default: 10) Maximum number of iterations in sequence of regressions
    norm : int
        (Optional, default: 2) Give the norm with which to normalize data
    constraints : list of function handles
        (Optional, default: None) A list of function handles for each variable that take the set of library terms and
        adds some sort of regulation
    scale_threshold : boolean
        (Optional, default: False) Whether to scale the threshold based on the L2-norm of the corresponding library
        function
    subset_threshold : list of dicts
        (Optional, default: None) List of dictionaries with a sub-sets of indices and a value for their threshold. Note
        That no overlap between the sub-sets is allowed. Any index not in a sub-set gets the threshold set by
        'threshold'.

    Returns
    -------
    np.array of weights for linear combination of library terms to best achieve data series

    """
    # Initialize weights and read the number of variables and the number of library terms
    w = []
    n_var = lib.shape[0]
    n_terms = lib.shape[1]

    if isinstance(threshold, float):
        tols = n_var * [threshold]
    elif isinstance(threshold, dict):
        try:
            tols = [threshold[key] for key in [f"t{n}" for n in range(n_var)]]
        except KeyError:
            raise KeyError("If thresholds are given as dictionary the keys need to be set as 't0', 't1' and so on.")
    else:
        tols = threshold

    assert len(tols) == n_var, "Thresholds need to be set for all variables. (Should be given as float for all" \
                               "or list-like.)"

    indices = set(np.arange(n_terms))
    sub_indices = []
    sub_thresholds = []
    if subset_threshold is not None:
        for subset in subset_threshold:
            sub_thresholds.append(subset["threshold"])
            sub_indices.append(set(subset["indices"]))
        _assert_no_intersections(sub_indices)
        for sub_index in sub_indices:
            indices = indices - sub_index

    # We try to find a set of weights for every variable's right hand side
    for var in range(n_var):
        # Read and reshape the library terms for the current variable, here we flatten all except the first dimension
        x = np.reshape(lib[var], (lib[var].shape[0], np.prod(lib[var].shape[1:])))
        # Flatten the corresponding data time series
        y = data[var].flatten()
        # When normalization is set, save the scaling factor and normalize the library according to the parameter norm
        l_norm = None
        tol_loc = tols[var]
        if scale_threshold:
            tol_scale = np.array([np.linalg.norm(item, 2) for item in x])
        else:
            tol_scale = np.ones(x.shape[0])

        if norm != 0:
            l_norm = 1.0 / (np.linalg.norm(x, norm))
            x = l_norm * x
            tol_loc = tol_loc / l_norm
        if lamb != 0 and constraints is None:
            # If a sparsity knob is given get the standard ridge estimate
            w_loc = np.linalg.lstsq(x.dot(x.T) + lamb * np.eye(n_terms), x.dot(y), rcond=None)[0]
        elif constraints is not None:
            w_loc = np.linalg.lstsq(x.dot(x.T) + lamb * np.eye(n_terms) +
                                    constraints[var](lib[var]), x.dot(y), rcond=None)[0]
        else:
            # else perform ordinary least squares fit
            w_loc = np.linalg.lstsq(x.T, y, rcond=None)[0]

        # Set the number of relevant terms to the number of all terms
        n_relevant = n_terms
        for i in range(max_it):
            # We find all the indices which hold values below our threshold and the remaining indices
            off_indices = np.where(np.abs(w_loc[list(indices)]) < tol_loc / tol_scale[list(indices)])[0]
            off_indices = set(np.array(list(indices))[off_indices])
            # TODO: make this such that sub_thresholds and indices can be set per variable
            # also I shouldn't have started with this set stuff
            off_sub_indices = set()
            for i in range(len(sub_indices)):
                off_sub = np.where(np.abs(w_loc)[list(sub_indices[i])] <
                                   sub_thresholds[i] / tol_scale[list(sub_indices[i])])[0]
                off_sub = set(np.array(list(sub_indices[i]))[off_sub])
                off_sub_indices.update(off_sub)
            all_off_indices = off_indices.union(off_sub_indices)
            new_on_indices = set(np.arange(n_terms)) - all_off_indices

            if n_relevant == len(new_on_indices):
                # Stop if the number of relevant terms does not change
                break
            else:
                # Update the number of relevant terms
                n_relevant = len(new_on_indices)

            on_indices = list(new_on_indices)

            if len(on_indices) == 0:
                # If all indices have been tossed stop (if this happened during the first iteration return the value)
                if i == 0:
                    # w.append(w_loc)
                    break
                else:
                    break

            # Cut off all values to the corresponding indices
            w_loc[list(all_off_indices)] = 0
            # Get new ridge or least square estimate
            if lamb != 0 and constraints is None:
                w_loc[on_indices] = np.linalg.lstsq(x[on_indices].dot(x[on_indices].T) +
                                                    lamb * np.eye(len(on_indices)), x[on_indices].dot(y), rcond=None)[0]
            elif constraints is not None:
                w_loc[on_indices] = np.linalg.lstsq(x[on_indices].dot(x[on_indices].T) +
                                                    lamb * np.eye(len(on_indices)) +
                                                    constraints[var](lib[var][on_indices]), x[on_indices].dot(y),
                                                    rcond=None)[0]
            else:
                w_loc[on_indices] = np.linalg.lstsq(x[on_indices].T, y, rcond=None)[0]

        # Normalize the weights
        if norm != 0:
            # w_loc[np.where(np.abs(w_loc) < tol_loc)[0]] = 0
            w.append(np.multiply(l_norm, w_loc))
        else:
            # w_loc[np.where(np.abs(w_loc) < tol_loc)[0]] = 0
            w.append(w_loc)

    return np.array(w)


def train_ridge(lib, data, lamb, thres, l0=None, split=0.7, seed=1234, max_it_train=10, max_it=10, norm=2):
    np.random.seed(seed)
    n_timesteps = lib.shape[2]
    x = np.reshape(lib, (lib.shape[0], lib.shape[1], np.prod(lib.shape[2:])))
    y = np.reshape(data, (data.shape[0], np.prod(data.shape[1:])))

    train_indices = np.random.choice(x.shape[-1], int(x.shape[-1] * split), replace=False)
    test_indices = np.delete(np.arange(x.shape[-1]), train_indices)

    x_train = x[:, :, train_indices]
    y_train = y[:, train_indices]

    x_test = x[:, :, test_indices]
    y_test = y[:, test_indices]

    d_tol = float(thres)
    tol = thres
    tol_best = tol

    if l0 is None:
        l0 = 0.001 * np.linalg.norm(np.linalg.cond(x), 2)

    def error(w_test):
        errors = [np.linalg.norm(y_test[i] - x_test[i].T.dot(w_test[i]), 2)
                  for i in range(x_test.shape[0])]
        error_test = np.sum(errors) + l0 * np.count_nonzero(w_test)
        return error_test

    w_best = np.array([np.linalg.lstsq(x_test[i].T, y_test[i], rcond=None)[0]
                       for i in range(x_test.shape[0])])
    error_best = error(w_best)

    pbar = tqdm(range(max_it_train))
    pbar.set_description(f"Training sequencing threshold ridge regression with "
                         f"{split * 100:.0f} % split")
    for i in pbar:
        w = sequencing_threshold_ridge(x_train, y_train, lamb, max_it=max_it, threshold=tol, norm=norm)
        err = error(w)

        if err <= error_best:
            error_best = err
            tol_best = tol
            w_best = w
            tol = tol + d_tol
        else:
            tol = max([0, tol - 2 * d_tol])
            d_tol = 2 * d_tol / (max_it_train - i)
            tol = tol + d_tol
        pbar.set_postfix({"Threshold": tol_best, "Error": error_best})

    return np.array(w_best), error_best


def train_ridge_wip(lib, data, data_dot, time, lamb, thres, l0=None, seed=1234, max_it_train=10, max_it=10, norm=2,
                    error="BIC"):
    np.random.seed(seed)
    n_timesteps = lib.shape[2]
    # reshaping is done inconsistently
    # x = np.reshape(lib, (lib.shape[0], lib.shape[1], np.prod(lib.shape[2:])))
    # y = np.reshape(data_dot, (data_dot.shape[0], np.prod(data_dot.shape[1:])))
    x = lib
    y = data_dot

    d_tol = thres
    tol = thres
    tol_best = tol

    if l0 is None:
        l0 = 0.001 * np.linalg.norm(np.linalg.cond(x.reshape((x.shape[0], x.shape[1], np.prod(x.shape[2:])))), 2)

    def error_der(w_test):
        errors = [np.linalg.norm(data_dot[i] - x[i].T.dot(w_test[i]).T) for i in range(x.shape[0])]
        error_test = np.sum(errors) + l0 * np.count_nonzero(w_test)
        return error_test

    def error_sp(w_test):
        errors = [np.linalg.norm(data[i, 1:] - cumtrapz(x[i].T.dot(w_test[i]), time).T)
                  for i in range(x.shape[0])]

        error_test = np.sum(errors) + l0 * np.count_nonzero(w_test)
        return error_test

    def error_bic(w_test):
        errors = np.array([np.linalg.norm(data[i, 1:] - cumtrapz(x[i].T.dot(w_test[i]), time).T)
                           for i in range(x.shape[0])])
        errors = errors.flatten() ** 2
        logL = -len(time) * np.log(errors.sum() / len(time)) / 2
        p = np.count_nonzero(w_test)
        bic = -2 * logL + p * np.log(len(time))
        return bic

    if error == "BIC":
        error_func = error_bic
    elif error == "integrate":
        error_func = error_sp
    elif error == "SINDy":
        error_func = error_der
    else:
        raise NotImplementedError("Unrecognized Error function")

    w_best = np.array([np.linalg.lstsq(x[i].reshape(x.shape[1], -1).T, y[i].flatten(), rcond=None)[0]
                       for i in range(x.shape[0])])
    error_best = error_func(w_best)

    pbar = tqdm(range(max_it_train))
    pbar.set_description(f"Training sequencing threshold ridge regression with " + error + "  error")
    for i in pbar:
        w = sequencing_threshold_ridge(x, y, lamb, max_it=max_it, threshold=tol, norm=norm)
        err = error_func(w)

        if err <= error_best:
            error_best = err
            tol_best = tol
            w_best = w
            tol = tol + d_tol
        else:
            tol = max([0, tol - 2 * d_tol])
            d_tol = 2 * d_tol / (max_it_train - i)
            tol = tol + d_tol
        pbar.set_postfix({"Threshold": tol_best, "Error": error_best})

    return np.array(w_best), error_best


def optimize_knob_and_threshold(lib, data, data_dot, time, max_evals=20, seed=420, **kwargs):
    def training_loss_bic(weights):
        errors = np.array([np.linalg.norm(data[i, 1:] - cumtrapz(lib[i].T.dot(weights[i]), time).T)
                           for i in range(lib.shape[0])])
        errors = errors.flatten() ** 2
        logL = -len(time) * np.log(errors.sum() / len(time)) / 2
        p = np.count_nonzero(weights)
        bic = -2 * logL  # + p * np.log(len(time))
        return (1 + 10 * (weights == 0.0).all(axis=1).sum()) * bic

    def BIC(args):
        lamb, thres = args
        weights = sequencing_threshold_ridge(lib, data_dot, lamb, threshold=thres, **kwargs)
        return training_loss_bic(weights)

    # s = hp.loguniform("l", -20, 0)
    # TODO: manage these with keyword arguments
    s = (hp.loguniform("l", -20, 0), hp.uniform("t", 0, 1.0))
    rstate = np.random.default_rng(seed)
    opt = fmin(BIC, space=s, algo=tpe.suggest, max_evals=max_evals, rstate=rstate)
    print(f"Optimal lambda found: {opt['l']}\n")
    print(f"Optimal threshold found: {opt['t']}")
    return sequencing_threshold_ridge(lib, data_dot, opt["l"], tol=opt["t"])


def optimize_threshold(lib, data, data_dot, time, max_evals=20, seed=420, space=None, condition=None,
                       loss=None, par_guess=0, subset_threshold=None, **kwargs):
    if condition is not None and loss is not None:
        raise UserWarning("Both a loss condition and a custom loss function are set; condition is being ignored")

    if condition is None:
        condition = lambda x, x_dot: 0

    if subset_threshold is None:
        subset_threshold = []

    def training_loss_bic(weights, lib, data, data_dot, time):
        errors = []
        conditions = []
        for var in range(lib.shape[0]):
            x_dot_recon = lib[var].T.dot(weights[var])
            x_recon = cumtrapz(x_dot_recon, time).T
            errors.append(np.linalg.norm(data[var, 1:] - x_recon))
            conditions.append(condition(x_recon, x_dot_recon))
        errors = np.array(errors)
        conditions = np.array(conditions)
        errors = errors.flatten() ** 2
        logL = -len(time) * np.log(errors.sum() / len(time)) / 2
        p = np.count_nonzero(weights)
        bic = -2 * logL + p * np.log(len(time)) + 200 * (par_guess != 0) * (p - par_guess) ** 2
        return (1 + 1000 * (weights == 0.0).all(axis=1).sum()) * bic * np.exp(conditions.sum() / len(time))

    # TODO: clean this up (also it's this ugly for backward compatibility)
    space_type = "uniform"
    space_bounds = [0.0, 1.0]
    if isinstance(space, str):
        space_type = space
        if space_type == "log":
            space_bounds = [-20, 0]
    elif isinstance(space, dict):
        space_type = space["type"]
        space_bounds = space["bounds"]

    if space_type == "uniform":
        s = lambda key: hp.uniform(key, space_bounds[0], space_bounds[1])
    elif space_type == "log":
        s = lambda key: hp.loguniform(key, space_bounds[0], space_bounds[1])

    rstate = np.random.default_rng(seed)
    thr_keys = [f"t{i}" for i in range(data.shape[0])]
    sub_thr_keys = [f"st{i}" for i in range(len(subset_threshold))]
    s_dict = {key: s(key) for key in thr_keys + sub_thr_keys}

    if loss is None:
        def training_loss(weights):
            return training_loss_bic(weights, lib, data, data_dot, time)
    else:
        def training_loss(weights):
            return loss(weights, lib, data, data_dot, time)

    def BIC(args):
        thres = [args[key] for key in thr_keys]
        sub_thres = [args[key] for key in sub_thr_keys]
        if len(sub_thres) != 0:
            sub_dict = [{"threshold": sub_thres[i], "indices": subset_threshold[i]} for i in range(len(sub_thres))]
        else:
            sub_dict = None
        weights = sequencing_threshold_ridge(lib, data_dot, 0.0, threshold=thres,
                                             subset_threshold=sub_dict, **kwargs)
        return training_loss(weights)

    opt = fmin(BIC, space=s_dict, algo=tpe.suggest, max_evals=max_evals, rstate=rstate)
    thresholds = [opt[key] for key in thr_keys]
    print(f"Optimal threshold(s) found: { {key: opt[key] for key in thr_keys} }")
    if len(sub_thr_keys) > 0:
        sub_thresholds = [opt[key] for key in sub_thr_keys]
        print(f"Optimal sub-threshold(s) found: { {key: opt[key] for key in sub_thr_keys} }")
        sub_dict = [{"threshold": sub_thresholds[i], "indices": subset_threshold[i]} for i in
                    range(len(sub_thresholds))]
    else:
        sub_dict = None

    return opt, sequencing_threshold_ridge(lib, data_dot, 0.0, threshold=thresholds, subset_threshold=sub_dict)


def optimize_threshold_with_delay(estimated_model, max_evals=20, seed=420, space=None, condition=None,
                                  loss=None, par_guess=0, **kwargs):
    # TODO: merge with 'optimize_threshold'
    lib = estimated_model.lot
    data = estimated_model.sol[:, :-1]
    data_dot = estimated_model.sol_dot
    time = estimated_model.time[:-1]

    if condition is not None and loss is not None:
        raise UserWarning("Both a loss condition and a custom loss function are set; condition is being ignored")

    if condition is None:
        condition = lambda x, x_dot: 0

    def training_loss_bic(weights, lib, data, data_dot, time):
        errors = []
        conditions = []
        for var in range(lib.shape[0]):
            x_dot_recon = lib[var].T.dot(weights[var])
            x_recon = cumtrapz(x_dot_recon, time).T
            errors.append(np.linalg.norm(data[var, 1:] - x_recon))
            conditions.append(condition(x_recon, x_dot_recon))
        errors = np.array(errors)
        conditions = np.array(conditions)
        errors = errors.flatten() ** 2
        logL = -len(time) * np.log(errors.sum() / len(time)) / 2
        p = np.count_nonzero(weights)
        bic = -2 * logL + p * np.log(len(time)) + 200 * (par_guess != 0) * (p - par_guess) ** 2
        # aic = -2 * logL + 2* p + 200 * (par_guess != 0) * (p - par_guess) ** 2
        return (1 + 1000 * (weights == 0.0).all(axis=1).sum()) * bic * np.exp(conditions.sum() / len(time))

    # TODO: clean this up (also it's this ugly for backward compatibility)
    space_type = "uniform"
    space_bounds = [0.0, 1.0]
    if isinstance(space, str):
        space_type = space
        if space_type == "log":
            space_bounds = [-20, 0]
    elif isinstance(space, dict):
        space_type = space["type"]
        space_bounds = space["bounds"]

    if space_type == "uniform":
        s = lambda key: hp.uniform(key, space_bounds[0], space_bounds[1])
    elif space_type == "log":
        s = lambda key: hp.loguniform(key, space_bounds[0], space_bounds[1])

    rstate = np.random.default_rng(seed)
    thr_keys = [f"t{i}" for i in range(data.shape[0])]
    s_dict = {key: s(key) for key in thr_keys}

    delay_type = "uniform"
    delay_bounds = [0.0, 2.0]
    if "delay" in space:
        delay_type = space["delay"]["type"]
        delay_bounds = space["delay"]["bounds"]

    if delay_type == "uniform":
        d = hp.uniform("delay", delay_bounds[0], delay_bounds[1])
    elif delay_type == "log":
        d = hp.loguniform("delay", delay_bounds[0], delay_bounds[1])
    elif delay_type == "randint":
        d = hp.randint("delay", delay_bounds[0], delay_bounds[1])

    s_dict["delay"] = d

    if loss is None:
        def training_loss(weights):
            return training_loss_bic(weights, lib, data, data_dot, time)
    else:
        def training_loss(weights):
            return loss(weights, lib, data, data_dot, time)

    def BIC(args):
        thres = [args[key] for key in thr_keys]
        delay = args["delay"]
        estimated_model.delay = delay
        estimated_model._recompile_delay_terms()
        weights = sequencing_threshold_ridge(lib, data_dot, 0.0, threshold=thres, **kwargs)
        return training_loss(weights)

    opt = fmin(BIC, space=s_dict, algo=tpe.suggest, max_evals=max_evals, rstate=rstate)
    thresholds = [opt[key] for key in thr_keys]
    print(f"Optimal threshold(s) found: { {key: opt[key] for key in thr_keys} }")
    print(f"Optimal delay found: {opt['delay']}")
    estimated_model.delay = opt["delay"]
    estimated_model._recompile_delay_terms()
    return opt, sequencing_threshold_ridge(lib, data_dot, 0.0, threshold=thresholds)


def optimize_threshold_optuna(lib, data, data_dot, time, max_evals=20, seed=420,
                              space="uniform", boundaries=None,
                              condition=None, loss=None, **kwargs):
    if condition is not None and loss is not None:
        raise UserWarning("Both a loss condition and a custom loss function are set; condition is being ignored")
    if condition is None:
        condition = lambda x, x_dot: 0
    if boundaries is None:
        boundaries = [1e-26, 1]

    def training_loss_bic(weights, lib, data, data_dot, time):
        errors = []
        conditions = []
        for var in range(lib.shape[0]):
            x_dot_recon = lib[var].T.dot(weights[var])
            x_recon = cumtrapz(x_dot_recon, time).T
            errors.append(np.linalg.norm(data[var, 1:] - x_recon))
            conditions.append(condition(x_recon, x_dot_recon))
        errors = np.array(errors)
        conditions = np.array(conditions)
        errors = errors.flatten() ** 2
        logL = -len(time) * np.log(errors.sum() / len(time)) / 2
        p = np.count_nonzero(weights)
        bic = -2 * logL + p * np.log(len(time))
        return (1 + 1000 * (weights == 0.0).all(axis=1).sum()) * bic * np.exp(conditions.sum() / len(time))
        # return bic * np.exp(conditions.sum() / len(time))

    thr_keys = [f"t{i}" for i in range(data.shape[0])]

    if space == "uniform":
        log = False
    else:
        log = True

    if loss is None:
        def training_loss(weights):
            return training_loss_bic(weights, lib, data, data_dot, time)
    else:
        def training_loss(weights):
            return loss(weights, lib, data, data_dot, time)

    def BIC(trial):
        thres = [trial.suggest_float(key, boundaries[0], boundaries[1], log=log) for key in thr_keys]
        # thres = [np.exp(trial.suggest_float(key, -30.0, 0.0, log=False)) for key in thr_keys]
        weights = sequencing_threshold_ridge(lib, data_dot, 0.0, threshold=thres, **kwargs)
        return training_loss(weights)

    study = optuna.create_study(sampler=TPESampler(seed=seed))
    study.optimize(BIC, n_trials=max_evals)
    thresholds = study.best_params
    print(f"Optimal threshold(s) found: {thresholds}")
    return study, sequencing_threshold_ridge(lib, data_dot, 0.0, threshold=thresholds, **kwargs)


def optimize_ridge_gfo(lib, data, split=0.7, seed=1234, trials=20, tol_range=[1e-4, 2], lamb_range=[0, 2]):
    np.random.seed(seed)
    # l0 = 0.001 * np.linalg.norm(np.linalg.cond(lib), 2)
    x = np.reshape(lib, (lib.shape[0], lib.shape[1], np.prod(lib.shape[2:])))
    y = np.reshape(data, (data.shape[0], np.prod(data.shape[1:])))

    train_indices = np.random.choice(x.shape[-1], int(x.shape[-1] * split), replace=False)
    test_indices = np.delete(np.arange(x.shape[-1]), train_indices)

    x_train = x[:, :, train_indices]
    y_train = y[:, train_indices]

    x_test = x[:, :, test_indices]
    y_test = y[:, test_indices]

    l0 = 0.001 * np.linalg.norm(np.linalg.cond(x), 2)

    def error(w_test):
        errors = [np.linalg.norm(y_test[i] - x_test[i].T.dot(w_test[i]), 2)
                  for i in range(x_test.shape[0])]
        error_test = np.sum(errors) + l0 * np.count_nonzero(w_test)
        return error_test

    def objective(para):
        tol = para["tol"]
        # tol = 1.0
        lamb = para["lamb"]
        # lamb = 0.01
        w = sequencing_threshold_ridge(x_train, y_train, lamb, tol=tol)
        # errors = [np.linalg.norm(data[i].flatten() - lib[i].T.dot(w[i]).flatten(), 2)
        #          for i in range(lib.shape[0])]
        # error_test = np.sum(errors) + l0 * np.count_nonzero(w)# + 1 / np.sum([np.linalg.norm(wi, 2) for wi in w])
        error_test = error(w)
        return - error_test

    search_space = {"tol": tol_range, "lamb": lamb_range}
    opt = HillClimbingOptimizer(search_space)
    opt.search(objective, n_iter=trials)

    lam_best, tol_best = opt.best_value

    return sequencing_threshold_ridge(lib, data, lam_best, tol=tol_best)


def optimize_wip(model, w_init, trials=20, n_time=5):
    time = model.simulator.time[:n_time]
    Nt = len(time)

    def error(w_test):
        # errors = [np.linalg.norm(y_test[i] - x_test[i].T.dot(w_test[i]), 2)
        #           for i in range(x_test.shape[0])]

        model.simulator.sigma = w_test
        y_estimate = model.simulator.simulate([time[0], time[-1]], t_eval=time)
        try:
            diff = ((model.sol - y_estimate).flatten()) ** 2
        except ValueError:
            diff = np.array([1e10])

        logL = -Nt * np.log(diff.sum() / Nt) / 2

        s = w_test.flatten()
        s[np.abs(s != 0)] /= np.abs(s[s != 0])
        p = np.abs(s).sum()

        bic = -2 * logL + p * np.log(Nt)

        return bic

    instru = ng.p.Instrumentation(ng.p.Array(init=w_init))
    optimizer = ng.optimizers.NGOpt(parametrization=instru, budget=trials)
    recommendation = optimizer.minimize(error)

    return recommendation.value[0][0]



"""
This module implements ddeint, a simple Differential Delay Equation
solver built on top of Scipy's odeint """

# REQUIRES Numpy and Scipy.
import numpy as np
import scipy.integrate
import scipy.interpolate


class ddeVar:
    """
    The instances of this class are special function-like
    variables which store their past values in an interpolator and
    can be called for any past time: Y(t), Y(t-d).
    Very convenient for the integration of DDEs.
    """

    def __init__(self, g, tc=0):
        """ g(t) = expression of Y(t) for t<tc """

        self.g = g
        self.tc = tc
        # We must fill the interpolator with 2 points minimum

        self.interpolator = scipy.interpolate.interp1d(
            np.array([tc - 1, tc]),  # X
            np.array([self.g(tc), self.g(tc)]).T,  # Y
            kind="linear",
            bounds_error=False,
            fill_value=self.g(tc)
        )

    def update(self, t, Y):
        """ Add one new (ti,yi) to the interpolator """
        # Y2 = Y if (Y.size == 1) else np.array([Y]).T
        Y2 = np.array([Y]).T
        self.interpolator = scipy.interpolate.interp1d(
            np.hstack([self.interpolator.x, [t]]),  # X
            np.hstack([self.interpolator.y, Y2]),  # Y
            kind="linear",
            bounds_error=False,
            fill_value=Y
        )

    def __call__(self, t=0):
        """ Y(t) will return the instance's value at time t """

        return self.g(t) if (t <= self.tc) else self.interpolator(t)


class dde(scipy.integrate.ode):
    """
    This class overwrites a few functions of ``scipy.integrate.ode``
    to allow for updates of the pseudo-variable Y between each
    integration step.
    """

    def __init__(self, f, jac=None):
        def f2(t, y, args):
            return f(self.Y, t, *args)

        scipy.integrate.ode.__init__(self, f2, jac)
        self.set_f_params(None)

    def integrate(self, t, step=0, relax=0):

        scipy.integrate.ode.integrate(self, t, step, relax)
        self.Y.update(self.t, self.y)
        return self.y

    def set_initial_value(self, Y):

        self.Y = Y  #!!! Y will be modified during integration
        scipy.integrate.ode.set_initial_value(self, Y(Y.tc), Y.tc)


def ddeint(func, g, tt, fargs=None):
    """ Solves Delay Differential Equations

    Similar to scipy.integrate.odeint. Solves a Delay differential
    Equation system (DDE) defined by

        Y(t) = g(t) for t<0
        Y'(t) = func(Y,t) for t>= 0

    Where func can involve past values of Y, like Y(t-d).
    

    Parameters
    -----------
    
    func
      a function Y,t,args -> Y'(t), where args is optional.
      The variable Y is an instance of class ddeVar, which means that
      it is called like a function: Y(t), Y(t-d), etc. Y(t) returns
      either a number or a numpy array (for multivariate systems).

    g
      The 'history function'. A function g(t)=Y(t) for t<0, g(t)
      returns either a number or a numpy array (for multivariate
      systems).
    
    tt
      The vector of times [t0, t1, ...] at which the system must
      be solved.

    fargs
      Additional arguments to be passed to parameter ``func``, if any.


    Examples
    ---------
    
    We will solve the delayed Lotka-Volterra system defined as
    
        For t < 0:
        x(t) = 1+t
        y(t) = 2-t
    
        For t >= 0:
        dx/dt =  0.5* ( 1- y(t-d) )
        dy/dt = -0.5* ( 1- x(t-d) )
    
    The delay ``d`` is a tunable parameter of the model.

    >>> import numpy as np
    >>> from ddeint import ddeint
    >>> 
    >>> def model(XY,t,d):
    >>>     x, y = XY(t)
    >>>     xd, yd = XY(t-d)
    >>>     return np.array([0.5*x*(1-yd), -0.5*y*(1-xd)])
    >>> 
    >>> g = lambda t : np.array([1+t,2-t]) # 'history' at t<0
    >>> tt = np.linspace(0,30,20000) # times for integration
    >>> d = 0.5 # set parameter d 
    >>> yy = ddeint(model,g,tt,fargs=(d,)) # solve the DDE !
     
    """

    dde_ = dde(func)
    dde_.set_initial_value(ddeVar(g, tt[0]))
    dde_.set_f_params(fargs if fargs else [])
    results = [dde_.integrate(dde_.t + dt) for dt in np.diff(tt)]
    return np.array([g(tt[0])] + results)
